@@ -1,26 +1,6 @@
-/* Simple Plugin API
- *
- * Copyright © 2018 Wim Taymans
- *
- * Permission is hereby granted, free of charge, to any person obtaining a
- * copy of this software and associated documentation files (the "Software"),
- * to deal in the Software without restriction, including without limitation
- * the rights to use, copy, modify, merge, publish, distribute, sublicense,
- * and/or sell copies of the Software, and to permit persons to whom the
- * Software is furnished to do so, subject to the following conditions:
- *
- * The above copyright notice and this permission notice (including the next
- * paragraph) shall be included in all copies or substantial portions of the
- * Software.
- *
- * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
- * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
- * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
- * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
- * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
- * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
- * DEALINGS IN THE SOFTWARE.
- */
+/* Simple Plugin API */
+/* SPDX-FileCopyrightText: Copyright © 2018 Wim Taymans */
+/* SPDX-License-Identifier: MIT */
 
 #ifndef SPA_NODE_H
 #define SPA_NODE_H
@@ -85,7 +65,7 @@ struct spa_node_info {
 	uint32_t n_params;			/**< number of items in \a params */
 };
 
-#define SPA_NODE_INFO_INIT()	(struct spa_node_info) { 0, }
+#define SPA_NODE_INFO_INIT()	((struct spa_node_info) { 0, })
 
 /**
  * Port information structure
@@ -124,7 +104,7 @@ struct spa_port_info {
 	uint32_t n_params;			/**< number of items in \a params */
 };
 
-#define SPA_PORT_INFO_INIT()	(struct spa_port_info) { 0, }
+#define SPA_PORT_INFO_INIT()	((struct spa_port_info) { 0, })
 
 #define SPA_RESULT_TYPE_NODE_ERROR	1
 #define SPA_RESULT_TYPE_NODE_PARAMS	2
@@ -470,6 +450,9 @@ struct spa_node_methods {
 	 * Enumerate all possible parameters of \a id on \a port_id of \a node
 	 * that are compatible with \a filter.
 	 *
+	 * When SPA_ID_INVALID is given as the port_id, the node will reply with
+	 * the params that would be returned for a new port in the given direction.
+	 *
 	 * The result parameters can be queried and modified and ultimately be used
 	 * to call port_set_param.
 	 *
@@ -484,7 +467,7 @@ struct spa_node_methods {
 	 * \param seq a sequence number to pass to the result event when
 	 *	this method is executed synchronously.
 	 * \param direction an spa_direction
-	 * \param port_id the port to query
+	 * \param port_id the port to query or SPA_ID_INVALID
 	 * \param id the parameter id to query
 	 * \param start the first index to query, 0 to get the first item
 	 * \param max the maximum number of params to query
@@ -505,7 +488,9 @@ struct spa_node_methods {
 	 *
 	 * When \a param is NULL, the parameter will be unset.
 	 *
-	 * This function must be called from the main thread.
+	 * This function must be called from the main thread. The node muse be paused
+	 * or the port SPA_IO_Buffers area is NULL when this function is called with
+	 * a param that changes the processing state (like a format change).
 	 *
 	 * \param node a struct \ref spa_node
 	 * \param direction a enum \ref spa_direction
@@ -560,7 +545,8 @@ struct spa_node_methods {
 	 * When this function returns async, use the spa_node_sync operation to
 	 * wait for completion.
 	 *
-	 * This function must be called from the main thread.
+	 * This function must be called from the main thread. The node muse be paused
+	 * or the port SPA_IO_Buffers area is NULL when this function is called.
 	 *
 	 * \param object an object implementing the interface
 	 * \param direction a port direction
@@ -586,6 +572,11 @@ struct spa_node_methods {
 	 *
 	 * This function must be called from the main thread.
 	 *
+	 * This function can be called when the node is running and the node
+	 * must be prepared to handle changes in io areas while running. This
+	 * is normally done by synchronizing the port io updates with the
+	 * data processing loop.
+	 *
 	 * \param direction a spa_direction
 	 * \param port_id a port id
 	 * \param id the id of the io area, the available ids can be
@@ -632,6 +623,12 @@ struct spa_node_methods {
 	 *
 	 * When the node can accept new input in the next cycle, the
 	 * SPA_STATUS_NEED_DATA bit will be set.
+	 *
+	 * Note that the node might return SPA_STATUS_NEED_DATA even when
+	 * no input ports have this status. This means that the amount of
+	 * data still available on the input ports is likely not going to
+	 * be enough for the next cycle and the host might need to prefetch
+	 * data for the next cycle.
 	 */
 	int (*process) (void *object);
 };
@@ -646,6 +643,16 @@ struct spa_node_methods {
 	_res;								\
 })
 
+#define spa_node_method_fast(o,method,version,...)			\
+({									\
+	int _res;							\
+	struct spa_node *_n = o;					\
+	spa_interface_call_fast_res(&_n->iface,				\
+			struct spa_node_methods, _res,			\
+			method, version, ##__VA_ARGS__);		\
+	_res;								\
+})
+
 #define spa_node_add_listener(n,...)		spa_node_method(n, add_listener, 0, __VA_ARGS__)
 #define spa_node_set_callbacks(n,...)		spa_node_method(n, set_callbacks, 0, __VA_ARGS__)
 #define spa_node_sync(n,...)			spa_node_method(n, sync, 0, __VA_ARGS__)
@@ -661,7 +668,9 @@ struct spa_node_methods {
 #define spa_node_port_set_io(n,...)		spa_node_method(n, port_set_io, 0, __VA_ARGS__)
 
 #define spa_node_port_reuse_buffer(n,...)	spa_node_method(n, port_reuse_buffer, 0, __VA_ARGS__)
+#define spa_node_port_reuse_buffer_fast(n,...)	spa_node_method_fast(n, port_reuse_buffer, 0, __VA_ARGS__)
 #define spa_node_process(n)			spa_node_method(n, process, 0)
+#define spa_node_process_fast(n)		spa_node_method_fast(n, process, 0)
 
 /**
  * \}
