@@ -12,6 +12,7 @@
 #include <alsa/asoundlib.h>
 
 #include <spa/node/node.h>
+#include <spa/node/keys.h>
 #include <spa/utils/type.h>
 #include <spa/utils/keys.h>
 #include <spa/utils/names.h>
@@ -26,6 +27,7 @@
 #include <spa/param/param.h>
 #include <spa/pod/filter.h>
 #include <spa/pod/parser.h>
+#include <spa/pod/dynamic.h>
 #include <spa/debug/pod.h>
 #include <spa/debug/log.h>
 
@@ -129,12 +131,12 @@ static int emit_node(struct impl *this, struct acp_device *dev)
 	struct spa_dict_item *items;
 	const struct acp_dict_item *it;
 	uint32_t n_items, i;
-	char device_name[128], path[180], channels[16], ch[12], routes[16];
-	char card_index[16], *p;
+	char device_name[128], path[210], channels[16], ch[12], routes[16];
+	char card_index[16], card_name[64], *p;
 	char positions[SPA_AUDIO_MAX_CHANNELS * 12];
 	struct spa_device_object_info info;
 	struct acp_card *card = this->card;
-	const char *stream, *devstr;
+	const char *stream, *devstr, *card_id;
 
 	info = SPA_DEVICE_OBJECT_INFO_INIT();
 	info.type = SPA_TYPE_INTERFACE_Node;
@@ -149,10 +151,12 @@ static int emit_node(struct impl *this, struct acp_device *dev)
 
 	info.change_mask = SPA_DEVICE_OBJECT_CHANGE_MASK_PROPS;
 
-	items = alloca((dev->props.n_items + 8) * sizeof(*items));
+	items = alloca((dev->props.n_items + 9) * sizeof(*items));
 	n_items = 0;
 
 	snprintf(card_index, sizeof(card_index), "%d", card->index);
+	card_id = acp_dict_lookup(&card->props, "alsa.id");
+	snprintf(card_name, sizeof(card_name), "%s", card_id ? card_id : card_index);
 
 	devstr = dev->device_strings[0];
 	p = strstr(devstr, "%f");
@@ -163,13 +167,15 @@ static int emit_node(struct impl *this, struct acp_device *dev)
 	} else {
 		snprintf(device_name, sizeof(device_name), "%s", devstr);
 	}
-	snprintf(path, sizeof(path), "alsa:pcm:%s:%s:%s", card_index, device_name, stream);
+
+	snprintf(path, sizeof(path), "alsa:acp:%s:%d:%s", card_name, dev->index, stream);
 	items[n_items++] = SPA_DICT_ITEM_INIT(SPA_KEY_OBJECT_PATH, path);
 	items[n_items++] = SPA_DICT_ITEM_INIT(SPA_KEY_API_ALSA_PATH, device_name);
 	if (dev->flags & ACP_DEVICE_UCM_DEVICE)
 		items[n_items++] = SPA_DICT_ITEM_INIT(SPA_KEY_API_ALSA_OPEN_UCM, "true");
 	items[n_items++] = SPA_DICT_ITEM_INIT(SPA_KEY_API_ALSA_PCM_CARD, card_index);
 	items[n_items++] = SPA_DICT_ITEM_INIT(SPA_KEY_API_ALSA_PCM_STREAM, stream);
+	items[n_items++] = SPA_DICT_ITEM_INIT(SPA_KEY_PORT_GROUP, stream);
 
 	snprintf(channels, sizeof(channels), "%d", dev->format.channels);
 	items[n_items++] = SPA_DICT_ITEM_INIT(SPA_KEY_AUDIO_CHANNELS, channels);
@@ -203,6 +209,7 @@ static int emit_info(struct impl *this, bool full)
 	struct acp_card *card = this->card;
 	char path[128];
 	uint64_t old = full ? this->info.change_mask : 0;
+	const char *card_id;
 
 	if (full)
 		this->info.change_mask = this->info_all;
@@ -210,11 +217,16 @@ static int emit_info(struct impl *this, bool full)
 		n_items = card->props.n_items + 4;
 		items = alloca(n_items * sizeof(*items));
 
+		card_id = acp_dict_lookup(&card->props, "alsa.id");
+
 		n_items = 0;
 #define ADD_ITEM(key, value) items[n_items++] = SPA_DICT_ITEM_INIT(key, value)
-		snprintf(path, sizeof(path), "alsa:pcm:%d", card->index);
+		if (card_id)
+			snprintf(path, sizeof(path), "alsa:acp:%s", card_id);
+		else
+			snprintf(path, sizeof(path), "alsa:acp:%d", card->index);
 		ADD_ITEM(SPA_KEY_OBJECT_PATH, path);
-		ADD_ITEM(SPA_KEY_DEVICE_API, "alsa:pcm");
+		ADD_ITEM(SPA_KEY_DEVICE_API, "alsa:acp");
 		ADD_ITEM(SPA_KEY_MEDIA_CLASS, "Audio/Device");
 		ADD_ITEM(SPA_KEY_API_ALSA_PATH,	(char *)this->props.device);
 		acp_dict_for_each(it, &card->props)
@@ -484,7 +496,8 @@ static int impl_enum_params(void *object, int seq,
 {
 	struct impl *this = object;
 	struct spa_pod *param;
-	struct spa_pod_builder b = { 0 };
+	spa_auto(spa_pod_dynamic_builder) b = { 0 };
+	struct spa_pod_builder_state state;
 	uint8_t buffer[4096];
 	struct spa_result_device_params result;
 	uint32_t count = 0;
@@ -496,6 +509,9 @@ static int impl_enum_params(void *object, int seq,
 	spa_return_val_if_fail(this != NULL, -EINVAL);
 	spa_return_val_if_fail(num != 0, -EINVAL);
 
+	spa_pod_dynamic_builder_init(&b, buffer, sizeof(buffer), 4096);
+	spa_pod_builder_get_state(&b.b, &state);
+
 	card = this->card;
 
 	result.id = id;
@@ -503,7 +519,7 @@ static int impl_enum_params(void *object, int seq,
       next:
 	result.index = result.next++;
 
-	spa_pod_builder_init(&b, buffer, sizeof(buffer));
+	spa_pod_builder_reset(&b.b, &state);
 
 	switch (id) {
 	case SPA_PARAM_EnumProfile:
@@ -511,15 +527,18 @@ static int impl_enum_params(void *object, int seq,
 			return 0;
 
 		pr = card->profiles[result.index];
-		param = build_profile(&b, id, pr, false);
+		if (SPA_FLAG_IS_SET(pr->flags, ACP_PROFILE_HIDDEN))
+			goto next;
+		param = build_profile(&b.b, id, pr, false);
 		break;
 
 	case SPA_PARAM_Profile:
 		if (result.index > 0 || card->active_profile_index >= card->n_profiles)
 			return 0;
-
 		pr = card->profiles[card->active_profile_index];
-		param = build_profile(&b, id, pr, true);
+		if (SPA_FLAG_IS_SET(pr->flags, ACP_PROFILE_HIDDEN))
+			goto next;
+		param = build_profile(&b.b, id, pr, true);
 		break;
 
 	case SPA_PARAM_EnumRoute:
@@ -527,7 +546,9 @@ static int impl_enum_params(void *object, int seq,
 			return 0;
 
 		p = card->ports[result.index];
-		param = build_route(&b, id, p, NULL, SPA_ID_INVALID);
+		if (SPA_FLAG_IS_SET(p->flags, ACP_PORT_HIDDEN))
+			goto next;
+		param = build_route(&b.b, id, p, NULL, SPA_ID_INVALID);
 		break;
 
 	case SPA_PARAM_Route:
@@ -536,6 +557,8 @@ static int impl_enum_params(void *object, int seq,
 				return 0;
 
 			dev = card->devices[result.index];
+			if (SPA_FLAG_IS_SET(dev->flags, ACP_DEVICE_HIDDEN))
+				goto next;
 			if (SPA_FLAG_IS_SET(dev->flags, ACP_DEVICE_ACTIVE) &&
 			    (p = find_port_for_device(card, dev)) != NULL)
 				break;
@@ -543,7 +566,9 @@ static int impl_enum_params(void *object, int seq,
 			result.index++;
 		}
 		result.next = result.index + 1;
-		param = build_route(&b, id, p, dev, card->active_profile_index);
+		if (SPA_FLAG_IS_SET(p->flags, ACP_PORT_HIDDEN))
+			goto next;
+		param = build_route(&b.b, id, p, dev, card->active_profile_index);
 		if (param == NULL)
 			return -errno;
 		break;
@@ -552,7 +577,7 @@ static int impl_enum_params(void *object, int seq,
 		return -ENOENT;
 	}
 
-	if (spa_pod_filter(&b, &result.param, param, filter) < 0)
+	if (spa_pod_filter(&b.b, &result.param, param, filter) < 0)
 		goto next;
 
 	spa_device_emit_result(&this->hooks, seq, 0,
@@ -694,6 +719,26 @@ static int apply_device_props(struct impl *this, struct acp_device *dev, struct
 	return changed;
 }
 
+static uint32_t find_profile_by_name(struct acp_card *card, const char *name)
+{
+	uint32_t i;
+	for (i = 0; i < card->n_profiles; i++) {
+		if (spa_streq(card->profiles[i]->name, name))
+			return i;
+	}
+	return SPA_ID_INVALID;
+}
+
+static uint32_t find_route_by_name(struct acp_card *card, const char *name)
+{
+	uint32_t i;
+	for (i = 0; i < card->n_ports; i++) {
+		if (spa_streq(card->ports[i]->name, name))
+			return i;
+	}
+	return SPA_ID_INVALID;
+}
+
 static int impl_set_param(void *object,
 			  uint32_t id, uint32_t flags,
 			  const struct spa_pod *param)
@@ -706,7 +751,8 @@ static int impl_set_param(void *object,
 	switch (id) {
 	case SPA_PARAM_Profile:
 	{
-		uint32_t idx;
+		uint32_t idx = SPA_ID_INVALID;
+		const char *name = NULL;
 		bool save = false;
 
 		if (param == NULL) {
@@ -714,20 +760,31 @@ static int impl_set_param(void *object,
 			save = true;
 		} else if ((res = spa_pod_parse_object(param,
 				SPA_TYPE_OBJECT_ParamProfile, NULL,
-				SPA_PARAM_PROFILE_index, SPA_POD_Int(&idx),
+				SPA_PARAM_PROFILE_index, SPA_POD_OPT_Int(&idx),
+				SPA_PARAM_PROFILE_name, SPA_POD_OPT_String(&name),
 				SPA_PARAM_PROFILE_save, SPA_POD_OPT_Bool(&save))) < 0) {
 			spa_log_warn(this->log, "can't parse profile");
 			spa_debug_log_pod(this->log, SPA_LOG_LEVEL_DEBUG, 0, NULL, param);
 			return res;
 		}
-
+		if (idx == SPA_ID_INVALID && name == NULL) {
+			spa_log_warn(this->log, "profile needs name or index");
+			return -EINVAL;
+		}
+		if (idx == SPA_ID_INVALID)
+			idx = find_profile_by_name(this->card, name);
+		if (idx == SPA_ID_INVALID) {
+			spa_log_warn(this->log, "unknown profile %s", name);
+			return -EINVAL;
+		}
 		acp_card_set_profile(this->card, idx, save ? ACP_PROFILE_SAVE : 0);
 		emit_info(this, false);
 		break;
 	}
 	case SPA_PARAM_Route:
 	{
-		uint32_t idx, device;
+		uint32_t idx = SPA_ID_INVALID, device;
+		const char *name = NULL;
 		struct spa_pod *props = NULL;
 		struct acp_device *dev;
 		bool save = false;
@@ -737,7 +794,8 @@ static int impl_set_param(void *object,
 
 		if ((res = spa_pod_parse_object(param,
 				SPA_TYPE_OBJECT_ParamRoute, NULL,
-				SPA_PARAM_ROUTE_index, SPA_POD_Int(&idx),
+				SPA_PARAM_ROUTE_index, SPA_POD_OPT_Int(&idx),
+				SPA_PARAM_ROUTE_name, SPA_POD_OPT_String(&name),
 				SPA_PARAM_ROUTE_device, SPA_POD_Int(&device),
 				SPA_PARAM_ROUTE_props, SPA_POD_OPT_Pod(&props),
 				SPA_PARAM_ROUTE_save, SPA_POD_OPT_Bool(&save))) < 0) {
@@ -747,8 +805,18 @@ static int impl_set_param(void *object,
 		}
 		if (device >= this->card->n_devices)
 			return -EINVAL;
+		if (idx == SPA_ID_INVALID && name == NULL)
+			return -EINVAL;
 
 		dev = this->card->devices[device];
+		if (SPA_FLAG_IS_SET(dev->flags, ACP_DEVICE_HIDDEN))
+			return -EINVAL;
+
+		if (idx == SPA_ID_INVALID)
+			idx = find_route_by_name(this->card, name);
+		if (idx == SPA_ID_INVALID)
+			return -EINVAL;
+
 		acp_device_set_port(dev, idx, save ? ACP_PORT_SAVE : 0);
 		if (props)
 			apply_device_props(this, dev, props);
