@@ -1,26 +1,6 @@
-/* PipeWire
- *
- * Copyright © 2021 Wim Taymans <wim.taymans@gmail.com>
- *
- * Permission is hereby granted, free of charge, to any person obtaining a
- * copy of this software and associated documentation files (the "Software"),
- * to deal in the Software without restriction, including without limitation
- * the rights to use, copy, modify, merge, publish, distribute, sublicense,
- * and/or sell copies of the Software, and to permit persons to whom the
- * Software is furnished to do so, subject to the following conditions:
- *
- * The above copyright notice and this permission notice (including the next
- * paragraph) shall be included in all copies or substantial portions of the
- * Software.
- *
- * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
- * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
- * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT.  IN NO EVENT SHALL
- * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
- * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
- * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
- * DEALINGS IN THE SOFTWARE.
- */
+/* PipeWire */
+/* SPDX-FileCopyrightText: Copyright © 2021 Wim Taymans <wim.taymans@gmail.com> */
+/* SPDX-License-Identifier: MIT */
 
 #include <unistd.h>
 #include <stdio.h>
@@ -36,6 +16,7 @@
 #include <spa/param/audio/format-utils.h>
 #include <spa/param/audio/raw.h>
 #include <spa/utils/json.h>
+#include <spa/debug/file.h>
 
 #include <pipewire/pipewire.h>
 #include <pipewire/impl.h>
@@ -85,7 +66,7 @@ static const struct pw_impl_module_events module_events = {
 
 static void show_help(struct data *data, const char *name, bool error)
 {
-        fprintf(error ? stderr : stdout, "%s [options]\n"
+	fprintf(error ? stderr : stdout, "%s [options]\n"
 		"  -h, --help                            Show this help\n"
 		"      --version                         Show version\n"
 		"  -r, --remote                          Remote daemon name\n"
@@ -131,6 +112,7 @@ int main(int argc, char *argv[])
 		{ NULL, 0, NULL, 0}
 	};
 	int c, res = -1;
+	struct spa_error_location loc;
 
 	setlocale(LC_ALL, "");
 	pw_init(&argc, &argv);
@@ -181,7 +163,7 @@ int main(int argc, char *argv[])
 			data.latency = atoi(optarg) * DEFAULT_RATE / SPA_MSEC_PER_SEC;
 			break;
 		case 'd':
-			data.delay = atof(optarg);
+			data.delay = (float)atof(optarg);
 			break;
 		case 'C':
 			pw_properties_set(data.capture_props, PW_KEY_TARGET_OBJECT, optarg);
@@ -190,10 +172,22 @@ int main(int argc, char *argv[])
 			pw_properties_set(data.playback_props, PW_KEY_TARGET_OBJECT, optarg);
 			break;
 		case 'i':
-			pw_properties_update_string(data.capture_props, optarg, strlen(optarg));
+			if (pw_properties_update_string_checked(data.capture_props,
+						optarg, strlen(optarg), &loc) < 0) {
+				spa_debug_file_error_location(stderr, &loc,
+						"error: syntax error in --capture-props: %s",
+						loc.reason);
+				return -1;
+			}
 			break;
 		case 'o':
-			pw_properties_update_string(data.playback_props, optarg, strlen(optarg));
+			if (pw_properties_update_string_checked(data.playback_props,
+						optarg, strlen(optarg), &loc) < 0) {
+				spa_debug_file_error_location(stderr, &loc,
+						"error: syntax error in --playback-props: %s",
+						loc.reason);
+				return -1;
+			}
 			break;
 		default:
 			show_help(&data, argv[0], true);
@@ -218,7 +212,7 @@ int main(int argc, char *argv[])
 	}
 
 
-        if ((f = open_memstream(&args, &size)) == NULL) {
+	if ((f = open_memstream(&args, &size)) == NULL) {
 		fprintf(stderr, "can't open memstream: %m\n");
 		goto exit;
 	}
